@@ -1,3 +1,24 @@
+/*
+ *
+ * Licensed to the Apache Software Foundation (ASF) under one
+ * or more contributor license agreements.  See the NOTICE file
+ * distributed with this work for additional information
+ * regarding copyright ownership.  The ASF licenses this file
+ * to you under the Apache License, Version 2.0 (the
+ * "License"); you may not use this file except in compliance
+ * with the License.  You may obtain a copy of the License at
+ *
+ *   http://www.apache.org/licenses/LICENSE-2.0
+ *
+ * Unless required by applicable law or agreed to in writing,
+ * software distributed under the License is distributed on an
+ * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ * KIND, either express or implied.  See the License for the
+ * specific language governing permissions and limitations
+ * under the License.
+ *
+ */
+
 #ifndef SVN_SWIG_SWIGUTIL_RB_H
 #define SVN_SWIG_SWIGUTIL_RB_H
 
@@ -35,6 +56,20 @@
 #  define SVN_RB_SWIG_SWIGUTIL_EXPORT
 #endif
 
+/* Ruby <=1.8.5 compatibility */
+#ifndef RARRAY_LEN
+#define RARRAY_LEN(x) RARRAY(x)->len
+#endif
+#ifndef RARRAY_PTR
+#define RARRAY_PTR(x) RARRAY(x)->ptr
+#endif
+#ifndef RSTRING_LEN
+#define RSTRING_LEN(x) RSTRING(x)->len
+#endif
+#ifndef RSTRING_PTR
+#define RSTRING_PTR(x) RSTRING(x)->ptr
+#endif
+
 #ifdef __cplusplus
 extern "C" {
 #endif /* __cplusplus */
@@ -52,6 +87,12 @@ typedef struct apr_pool_wrapper_t
 SVN_RB_SWIG_SWIGUTIL_EXPORT
 void svn_swig_rb_initialize(void);
 
+SVN_RB_SWIG_SWIGUTIL_EXPORT
+apr_pool_t *svn_swig_rb_pool(void);
+
+SVN_RB_SWIG_SWIGUTIL_EXPORT
+apr_allocator_t *svn_swig_rb_allocator(void);
+
 SVN_RB_SWIG_SWIGUTIL_EXPORT
 VALUE svn_swig_rb_svn_delta_editor(void);
 SVN_RB_SWIG_SWIGUTIL_EXPORT
@@ -64,7 +105,7 @@ void svn_swig_rb_raise_svn_repos_already_close(void);
 
 SVN_RB_SWIG_SWIGUTIL_EXPORT
 VALUE svn_swig_rb_svn_error_new(VALUE code, VALUE message,
-                                VALUE file, VALUE line);
+                                VALUE file, VALUE line, VALUE child);
 SVN_RB_SWIG_SWIGUTIL_EXPORT
 VALUE svn_swig_rb_svn_error_to_rb_error(svn_error_t *error);
 SVN_RB_SWIG_SWIGUTIL_EXPORT
