@@ -1,3 +1,22 @@
+# ====================================================================
+#    Licensed to the Apache Software Foundation (ASF) under one
+#    or more contributor license agreements.  See the NOTICE file
+#    distributed with this work for additional information
+#    regarding copyright ownership.  The ASF licenses this file
+#    to you under the Apache License, Version 2.0 (the
+#    "License"); you may not use this file except in compliance
+#    with the License.  You may obtain a copy of the License at
+#
+#      http://www.apache.org/licenses/LICENSE-2.0
+#
+#    Unless required by applicable law or agreed to in writing,
+#    software distributed under the License is distributed on an
+#    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+#    KIND, either express or implied.  See the License for the
+#    specific language governing permissions and limitations
+#    under the License.
+# ====================================================================
+
 require "English"
 require "time"
 require "stringio"
@@ -7,7 +26,7 @@ require "svn/error"
 require "svn/ext/core"
 
 class Time
-  MILLION = 1_000_000
+  MILLION = 1_000_000 #:nodoc:
 
   class << self
     def from_apr_time(apr_time)
@@ -84,6 +103,9 @@ module Svn
     Pool = Svn::Ext::Core::Apr_pool_wrapper_t
 
     class Pool
+      RECOMMENDED_MAX_FREE_SIZE = ALLOCATOR_RECOMMENDED_MAX_FREE
+      MAX_FREE_UNLIMITED = ALLOCATOR_MAX_FREE_UNLIMITED
+
       class << self
         def number_of_pools
           ObjectSpace.each_object(Pool) {}
@@ -187,9 +209,15 @@ module Svn
         add_provider(Core.auth_get_simple_provider)
       end
 
-      if Core.respond_to?(:auth_get_windows_simple_provider)
-        def add_windows_simple_provider
-          add_provider(Core.auth_get_windows_simple_provider)
+      if Util.windows?
+        if Core.respond_to?(:auth_get_windows_simple_provider)
+          def add_windows_simple_provider
+            add_provider(Core.auth_get_windows_simple_provider)
+          end
+        elsif Core.respond_to?(:auth_get_platform_specific_provider)
+          def add_windows_simple_provider
+            add_provider(Core.auth_get_platform_specific_provider("windows","simple"))
+          end
         end
       end
 
@@ -641,16 +669,16 @@ module Svn
 
       def infinity_or_empty_from_recurse(depth_or_recurse)
         case depth_or_recurse
-          when true  : DEPTH_INFINITY
-          when false : DEPTH_EMPTY
+          when true  then DEPTH_INFINITY
+          when false then DEPTH_EMPTY
           else depth_or_recurse
         end
       end
 
       def infinity_or_immediates_from_recurse(depth_or_recurse)
         case depth_or_recurse
-          when true  : DEPTH_INFINITY
-          when false : DEPTH_IMMEDIATES
+          when true  then DEPTH_INFINITY
+          when false then DEPTH_IMMEDIATES
           else depth_or_recurse
         end
       end
@@ -742,7 +770,7 @@ module Svn
         end
       end
 
-      def diff(to, consider_inheritance=nil)
+      def diff(to, consider_inheritance=false)
         Core.mergeinfo_diff(self, to, consider_inheritance).collect do |result|
           self.class.new(result)
         end
@@ -773,7 +801,7 @@ module Svn
         end
       end
 
-      def diff(to, consider_inheritance=nil)
+      def diff(to, consider_inheritance=false)
         result = Core.rangelist_diff(self, to, consider_inheritance)
         deleted = result.pop
         added = result
@@ -786,11 +814,12 @@ module Svn
         self.class.new(*Core.swig_rangelist_merge(self, changes))
       end
 
-      def remove(eraser, consider_inheritance=nil)
-        self.class.new(*Core.rangelist_remove(eraser, self, consider_inheritance))
+      def remove(eraser, consider_inheritance=false)
+        self.class.new(*Core.rangelist_remove(eraser, self,
+                                              consider_inheritance))
       end
 
-      def intersect(other, consider_inheritance=nil)
+      def intersect(other, consider_inheritance=false)
         self.class.new(*Core.rangelist_intersect(self, other,
                                                  consider_inheritance))
       end
