@@ -1,17 +1,22 @@
 /* reps-strings.c : intepreting representations with respect to strings
  *
  * ====================================================================
- * Copyright (c) 2000-2009 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
@@ -127,7 +132,7 @@ delta_string_keys(apr_array_header_t **keys,
 
 /* Delete the strings associated with array KEYS in FS as part of TRAIL.  */
 static svn_error_t *
-delete_strings(apr_array_header_t *keys,
+delete_strings(const apr_array_header_t *keys,
                svn_fs_t *fs,
                trail_t *trail,
                apr_pool_t *pool)
@@ -318,7 +323,7 @@ get_one_window(struct compose_handler_baton *cb,
 
 static svn_error_t *
 rep_undeltify_range(svn_fs_t *fs,
-                    apr_array_header_t *deltas,
+                    const apr_array_header_t *deltas,
                     representation_t *fulltext,
                     int cur_chunk,
                     char *buf,
@@ -528,7 +533,7 @@ rep_read_range(svn_fs_t *fs,
                    _("Corruption detected whilst reading delta chain from "
                      "representation '%s' to '%s'"), first_rep_key, rep_key);
               else
-                return err;
+                return svn_error_trace(err);
             }
         }
     }
@@ -818,13 +823,11 @@ svn_fs_base__rep_contents(svn_string_t *str,
                          pool));
 
     if (! svn_checksum_match(checksum, rep_checksum))
-      return svn_error_createf
-        (SVN_ERR_FS_CORRUPT, NULL,
-         _("Checksum mismatch on representation '%s':\n"
-           "   expected:  %s\n"
-           "     actual:  %s\n"), rep_key,
-         svn_checksum_to_cstring_display(rep_checksum, pool),
-         svn_checksum_to_cstring_display(checksum, pool));
+      return svn_error_create(SVN_ERR_FS_CORRUPT,
+                svn_checksum_mismatch_err(rep_checksum, checksum, pool,
+                            _("Checksum mismatch on representation '%s'"),
+                            rep_key),
+                NULL);
   }
 
   return SVN_NO_ERROR;
@@ -899,10 +902,12 @@ txn_body_read_rep(void *baton, trail_t *trail)
             {
               representation_t *rep;
 
-              svn_checksum_final(&args->rb->md5_checksum,
-                                 args->rb->md5_checksum_ctx, trail->pool);
-              svn_checksum_final(&args->rb->sha1_checksum,
-                                 args->rb->sha1_checksum_ctx, trail->pool);
+              SVN_ERR(svn_checksum_final(&args->rb->md5_checksum,
+                                         args->rb->md5_checksum_ctx,
+                                         trail->pool));
+              SVN_ERR(svn_checksum_final(&args->rb->sha1_checksum,
+                                         args->rb->sha1_checksum_ctx,
+                                         trail->pool));
               args->rb->checksum_finalized = TRUE;
 
               SVN_ERR(svn_fs_bdb__read_rep(&rep, args->rb->fs,
@@ -912,28 +917,22 @@ txn_body_read_rep(void *baton, trail_t *trail)
               if (rep->md5_checksum
                   && (! svn_checksum_match(rep->md5_checksum,
                                            args->rb->md5_checksum)))
-                return svn_error_createf
-                  (SVN_ERR_FS_CORRUPT, NULL,
-                   _("MD5 checksum mismatch on representation '%s':\n"
-                     "   expected:  %s\n"
-                     "     actual:  %s\n"), args->rb->rep_key,
-                   svn_checksum_to_cstring_display(rep->md5_checksum,
-                                                   trail->pool),
-                   svn_checksum_to_cstring_display(args->rb->md5_checksum,
-                                                   trail->pool));
+                return svn_error_create(SVN_ERR_FS_CORRUPT,
+                        svn_checksum_mismatch_err(rep->md5_checksum,
+                             args->rb->sha1_checksum, trail->pool,
+                             _("MD5 checksum mismatch on representation '%s'"),
+                             args->rb->rep_key),
+                        NULL);
 
               if (rep->sha1_checksum
                   && (! svn_checksum_match(rep->sha1_checksum,
                                            args->rb->sha1_checksum)))
-                return svn_error_createf
-                  (SVN_ERR_FS_CORRUPT, NULL,
-                   _("SHA1 checksum mismatch on representation '%s':\n"
-                     "   expected:  %s\n"
-                     "     actual:  %s\n"), args->rb->rep_key,
-                   svn_checksum_to_cstring_display(rep->sha1_checksum,
-                                                   trail->pool),
-                   svn_checksum_to_cstring_display(args->rb->sha1_checksum,
-                                                   trail->pool));
+                return svn_error_createf(SVN_ERR_FS_CORRUPT,
+                        svn_checksum_mismatch_err(rep->sha1_checksum,
+                            args->rb->sha1_checksum, trail->pool,
+                            _("SHA1 checksum mismatch on representation '%s'"),
+                            args->rb->rep_key),
+                        NULL);
             }
         }
     }
@@ -966,18 +965,16 @@ rep_read_contents(void *baton, char *buf, apr_size_t *len)
     SVN_ERR(txn_body_read_rep(&args, rb->trail));
   else
     {
-      /* Hey, guess what?  trails don't clear their own subpools.  In
-         the case of reading from the db, any returned data should
+      /* In the case of reading from the db, any returned data should
          live in our pre-allocated buffer, so the whole operation can
          happen within a single malloc/free cycle.  This prevents us
          from creating millions of unnecessary trail subpools when
-         reading a big file. */
-      apr_pool_t *subpool = svn_pool_create(rb->pool);
+         reading a big file.  */
       SVN_ERR(svn_fs_base__retry_txn(rb->fs,
                                      txn_body_read_rep,
                                      &args,
-                                     subpool));
-      svn_pool_destroy(subpool);
+                                     TRUE,
+                                     rb->pool));
     }
   return SVN_NO_ERROR;
 }
@@ -1136,19 +1133,17 @@ rep_write_contents(void *baton,
     SVN_ERR(txn_body_write_rep(&args, wb->trail));
   else
     {
-      /* Hey, guess what?  trails don't clear their own subpools.  In
-         the case of simply writing the rep to the db, we're *certain*
-         that there's no data coming back to us that needs to be
-         preserved... so the whole operation can happen within a
+      /* In the case of simply writing the rep to the db, we're
+         *certain* that there's no data coming back to us that needs
+         to be preserved... so the whole operation can happen within a
          single malloc/free cycle.  This prevents us from creating
          millions of unnecessary trail subpools when writing a big
          file. */
-      apr_pool_t *subpool = svn_pool_create(wb->pool);
       SVN_ERR(svn_fs_base__retry_txn(wb->fs,
                                      txn_body_write_rep,
                                      &args,
-                                     subpool));
-      svn_pool_destroy(subpool);
+                                     TRUE,
+                                     wb->pool));
     }
 
   return SVN_NO_ERROR;
@@ -1203,8 +1198,10 @@ rep_write_close_contents(void *baton)
   if (wb->trail)
     return txn_body_write_close_rep(wb, wb->trail);
   else
+    /* We need to keep our trail pool around this time so the
+       checksums we've calculated survive. */
     return svn_fs_base__retry_txn(wb->fs, txn_body_write_close_rep,
-                                  wb, wb->pool);
+                                  wb, FALSE, wb->pool);
 }
 
 
