@@ -2,17 +2,22 @@
  * locks.c :  entry point for locking RA functions for ra_serf
  *
  * ====================================================================
- * Copyright (c) 2006 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
@@ -24,14 +29,12 @@
 
 #include <serf.h>
 
+#include "svn_dav.h"
 #include "svn_pools.h"
 #include "svn_ra.h"
-#include "svn_dav.h"
-#include "svn_xml.h"
+
 #include "../libsvn_ra/ra_loader.h"
 #include "svn_config.h"
-#include "svn_delta.h"
-#include "svn_version.h"
 #include "svn_path.h"
 #include "svn_time.h"
 #include "svn_private_config.h"
@@ -42,7 +45,7 @@
 /*
  * This enum represents the current state of our XML parsing for a REPORT.
  */
-typedef enum {
+typedef enum lock_state_e {
   NONE = 0,
   PROP,
   LOCK_DISCOVERY,
@@ -55,12 +58,12 @@ typedef enum {
   COMMENT,
 } lock_state_e;
 
-typedef struct {
+typedef struct lock_prop_info_t {
   const char *data;
   apr_size_t len;
 } lock_prop_info_t;
 
-typedef struct {
+typedef struct lock_info_t {
   apr_pool_t *pool;
 
   const char *path;
@@ -82,9 +85,6 @@ typedef struct {
 
   /* are we done? */
   svn_boolean_t done;
-
-  /* Any errors. */
-  svn_error_t *error;
 } lock_info_t;
 
 
@@ -175,7 +175,7 @@ start_lock(svn_ra_serf__xml_parser_t *parser,
         }
       else
         {
-          abort();
+          SVN_ERR_MALFUNCTION();
         }
     }
   else if (state == LOCK_SCOPE)
@@ -186,7 +186,7 @@ start_lock(svn_ra_serf__xml_parser_t *parser,
         }
       else
         {
-          abort();
+          SVN_ERR_MALFUNCTION();
         }
     }
 
@@ -319,14 +319,14 @@ static const svn_ra_serf__dav_props_t lock_props[] =
   { NULL }
 };
 
-static apr_status_t
+static svn_error_t *
 set_lock_headers(serf_bucket_t *headers,
                  void *baton,
                  apr_pool_t *pool)
 {
   lock_info_t *lock_ctx = baton;
 
-  if (lock_ctx->force == TRUE)
+  if (lock_ctx->force)
     {
       serf_bucket_headers_set(headers, SVN_DAV_OPTIONS_HEADER,
                               SVN_DAV_OPTION_LOCK_STEAL);
@@ -341,7 +341,8 @@ set_lock_headers(serf_bucket_t *headers,
   return APR_SUCCESS;
 }
 
-static apr_status_t
+/* Implements svn_ra_serf__response_handler_t */
+static svn_error_t *
 handle_lock(serf_request_t *request,
             serf_bucket_t *response,
             void *handler_baton,
@@ -349,7 +350,7 @@ handle_lock(serf_request_t *request,
 {
   svn_ra_serf__xml_parser_t *xml_ctx = handler_baton;
   lock_info_t *ctx = xml_ctx->user_data;
-  apr_status_t status;
+  svn_error_t *err;
 
   if (ctx->read_headers == FALSE)
     {
@@ -357,9 +358,13 @@ handle_lock(serf_request_t *request,
       const char *val;
 
       serf_status_line sl;
-      apr_status_t rv;
+      apr_status_t status;
 
-      rv = serf_bucket_response_status(response, &sl);
+      status = serf_bucket_response_status(response, &sl);
+      if (SERF_BUCKET_READ_ERROR(status))
+        {
+          return svn_error_wrap_apr(status, NULL);
+        }
 
       ctx->status_code = sl.code;
       ctx->reason = sl.reason;
@@ -367,18 +372,17 @@ handle_lock(serf_request_t *request,
       /* 423 == Locked */
       if (sl.code == 423)
         {
-          ctx->error = svn_ra_serf__handle_server_error(request, response,
-                                                        pool);
-
-          /* Older servers may not give a descriptive error. */
-          if (!ctx->error)
+          /* Older servers may not give a descriptive error, so we'll
+             make one of our own if we can't find one in the response. */
+          err = svn_ra_serf__handle_server_error(request, response, pool);
+          if (!err)
             {
-              ctx->error = svn_error_createf(SVN_ERR_FS_PATH_ALREADY_LOCKED,
-                                             NULL,
-                                             _("Lock request failed: %d %s"),
-                                             ctx->status_code, ctx->reason);
+              err = svn_error_createf(SVN_ERR_FS_PATH_ALREADY_LOCKED,
+                                      NULL,
+                                      _("Lock request failed: %d %s"),
+                                      ctx->status_code, ctx->reason);
             }
-          return ctx->error->apr_err;
+          return err;
         }
 
       headers = serf_bucket_response_get_headers(response);
@@ -392,15 +396,8 @@ handle_lock(serf_request_t *request,
       val = serf_bucket_headers_get(headers, SVN_DAV_CREATIONDATE_HEADER);
       if (val)
         {
-          svn_error_t *err;
-
-          err = svn_time_from_cstring(&ctx->lock->creation_date, val,
-                                      ctx->pool);
-          if (err)
-            {
-              xml_ctx->error = err;
-              return err->apr_err;
-            }
+          SVN_ERR(svn_time_from_cstring(&ctx->lock->creation_date, val,
+                                        ctx->pool));
         }
 
       ctx->read_headers = TRUE;
@@ -409,78 +406,93 @@ handle_lock(serf_request_t *request,
   /* Forbidden when a lock doesn't exist. */
   if (ctx->status_code == 403)
     {
-      status = svn_ra_serf__handle_discard_body(request, response, NULL, pool);
-      if (APR_STATUS_IS_EOF(status))
+      /* If we get an "unexpected EOF" error, we'll wrap it with
+         generic request failure error. */
+      err = svn_ra_serf__handle_discard_body(request, response, NULL, pool);
+      if (err && APR_STATUS_IS_EOF(err->apr_err))
         {
           ctx->done = TRUE;
-          ctx->error = svn_error_createf(SVN_ERR_RA_DAV_REQUEST_FAILED, NULL,
-                                         _("Lock request failed: %d %s"),
-                                         ctx->status_code, ctx->reason);
+          err = svn_error_createf(SVN_ERR_RA_DAV_REQUEST_FAILED,
+                                  err,
+                                  _("Lock request failed: %d %s"),
+                                  ctx->status_code, ctx->reason);
         }
-    }
-  else
-    {
-      status = svn_ra_serf__handle_xml_parser(request, response, handler_baton,
-                                              pool);
+      return err;
     }
 
-  return status;
+  return svn_ra_serf__handle_xml_parser(request, response,
+                                        handler_baton, pool);
 }
 
-#define GET_LOCK "<?xml version=\"1.0\" encoding=\"utf-8\"?><propfind xmlns=\"DAV:\"><prop><lockdiscovery/></prop></propfind>"
-
-static serf_bucket_t*
-create_getlock_body(void *baton,
+/* Implements svn_ra_serf__request_body_delegate_t */
+static svn_error_t *
+create_getlock_body(serf_bucket_t **body_bkt,
+                    void *baton,
                     serf_bucket_alloc_t *alloc,
                     apr_pool_t *pool)
 {
-  serf_bucket_t *buckets, *tmp;
+  serf_bucket_t *buckets;
 
   buckets = serf_bucket_aggregate_create(alloc);
-  tmp = SERF_BUCKET_SIMPLE_STRING_LEN(GET_LOCK, sizeof(GET_LOCK)-1, alloc);
-  serf_bucket_aggregate_append(buckets, tmp);
 
-  return buckets;
+  svn_ra_serf__add_xml_header_buckets(buckets, alloc);
+  svn_ra_serf__add_open_tag_buckets(buckets, alloc, "propfind",
+                                    "xmlns", "DAV:",
+                                    NULL);
+  svn_ra_serf__add_open_tag_buckets(buckets, alloc, "prop", NULL);
+  svn_ra_serf__add_tag_buckets(buckets, "lockdiscovery", NULL, alloc);
+  svn_ra_serf__add_close_tag_buckets(buckets, alloc, "prop");
+  svn_ra_serf__add_close_tag_buckets(buckets, alloc, "propfind");
+
+  *body_bkt = buckets;
+  return SVN_NO_ERROR;
 }
 
-#define LOCK_HEADER "<?xml version=\"1.0\" encoding=\"utf-8\"?><lockinfo xmlns=\"DAV:\">"
-#define LOCK_TRAILER "</lockinfo>"
+static svn_error_t*
+setup_getlock_headers(serf_bucket_t *headers,
+                      void *baton,
+                      apr_pool_t *pool)
+{
+  serf_bucket_headers_set(headers, "Depth", "0");
 
-static serf_bucket_t*
-create_lock_body(void *baton,
+  return SVN_NO_ERROR;
+}
+
+/* Implements svn_ra_serf__request_body_delegate_t */
+static svn_error_t *
+create_lock_body(serf_bucket_t **body_bkt,
+                 void *baton,
                  serf_bucket_alloc_t *alloc,
                  apr_pool_t *pool)
 {
   lock_info_t *ctx = baton;
-  serf_bucket_t *buckets, *tmp;
+  serf_bucket_t *buckets;
 
   buckets = serf_bucket_aggregate_create(alloc);
 
-  tmp = SERF_BUCKET_SIMPLE_STRING_LEN(LOCK_HEADER, sizeof(LOCK_HEADER)-1,
-                                      alloc);
-  serf_bucket_aggregate_append(buckets, tmp);
+  svn_ra_serf__add_xml_header_buckets(buckets, alloc);
+  svn_ra_serf__add_open_tag_buckets(buckets, alloc, "lockinfo",
+                                    "xmlns", "DAV:",
+                                    NULL);
 
-  svn_ra_serf__add_tag_buckets(buckets, "lockscope", "<exclusive/>", alloc);
+  svn_ra_serf__add_open_tag_buckets(buckets, alloc, "lockscope", NULL);
+  svn_ra_serf__add_tag_buckets(buckets, "exclusive", NULL, alloc);
+  svn_ra_serf__add_close_tag_buckets(buckets, alloc, "lockscope");
 
-  svn_ra_serf__add_tag_buckets(buckets, "locktype", "<write/>", alloc);
+  svn_ra_serf__add_open_tag_buckets(buckets, alloc, "locktype", NULL);
+  svn_ra_serf__add_tag_buckets(buckets, "write", NULL, alloc);
+  svn_ra_serf__add_close_tag_buckets(buckets, alloc, "locktype");
 
   if (ctx->lock->comment)
     {
-      svn_stringbuf_t *xml_esc = NULL;
-      svn_string_t val;
-
-      val.data = ctx->lock->comment;
-      val.len = strlen(ctx->lock->comment);
-
-      svn_xml_escape_cdata_string(&xml_esc, &val, pool);
-      svn_ra_serf__add_tag_buckets(buckets, "owner", xml_esc->data, alloc);
+      svn_ra_serf__add_tag_buckets(buckets, "owner", ctx->lock->comment,
+                                   alloc);
     }
 
-  tmp = SERF_BUCKET_SIMPLE_STRING_LEN(LOCK_TRAILER, sizeof(LOCK_TRAILER)-1,
-                                      alloc);
-  serf_bucket_aggregate_append(buckets, tmp);
+  svn_ra_serf__add_close_tag_buckets(buckets, alloc, "lockinfo");
 
-  return buckets;
+  *body_bkt = buckets;
+  return SVN_NO_ERROR;
 }
 
 svn_error_t *
@@ -497,7 +509,7 @@ svn_ra_serf__get_lock(svn_ra_session_t *ra_session,
   svn_error_t *err;
   int status_code;
 
-  req_url = svn_path_url_add_component(session->repos_url.path, path, pool);
+  req_url = svn_path_url_add_component2(session->session_url.path, path, pool);
 
   lock_ctx = apr_pcalloc(pool, sizeof(*lock_ctx));
 
@@ -527,24 +539,18 @@ svn_ra_serf__get_lock(svn_ra_session_t *ra_session,
   handler->body_delegate = create_getlock_body;
   handler->body_delegate_baton = lock_ctx;
 
+  handler->header_delegate = setup_getlock_headers;
+  handler->header_delegate_baton = lock_ctx;
+
   handler->response_handler = handle_lock;
   handler->response_baton = parser_ctx;
 
   svn_ra_serf__request_create(handler);
   err = svn_ra_serf__context_run_wait(&lock_ctx->done, session, pool);
-  if (lock_ctx->error || parser_ctx->error)
-    {
-      svn_error_clear(err);
-      /* A 403 forbidden error indicates there's no lock, which we can ignore
-         here. */
-      if (lock_ctx->error && lock_ctx->status_code == 403)
-        svn_error_clear(lock_ctx->error);
-      SVN_ERR(parser_ctx->error);
-    }
 
   if (status_code == 404)
     {
-      return svn_error_create(SVN_ERR_RA_ILLEGAL_URL, NULL,
+      return svn_error_create(SVN_ERR_RA_ILLEGAL_URL, err,
                               _("Malformed URL for repository"));
     }
   if (err)
@@ -572,7 +578,7 @@ svn_ra_serf__lock(svn_ra_session_t *ra_session,
   apr_hash_index_t *hi;
   apr_pool_t *subpool;
 
-  apr_pool_create(&subpool, pool);
+  subpool = svn_pool_create(pool);
 
   for (hi = apr_hash_first(pool, path_revs); hi; hi = apr_hash_next(hi))
     {
@@ -583,6 +589,7 @@ svn_ra_serf__lock(svn_ra_session_t *ra_session,
       const void *key;
       void *val;
       svn_error_t *err;
+      svn_error_t *new_err = NULL;
 
       svn_pool_clear(subpool);
 
@@ -597,8 +604,8 @@ svn_ra_serf__lock(svn_ra_session_t *ra_session,
       lock_ctx->lock->comment = comment;
 
       lock_ctx->force = force;
-      req_url = svn_path_url_add_component(session->repos_url.path,
-                                           lock_ctx->path, subpool);
+      req_url = svn_path_url_add_component2(session->session_url.path,
+                                            lock_ctx->path, subpool);
 
       handler = apr_pcalloc(subpool, sizeof(*handler));
 
@@ -628,20 +635,13 @@ svn_ra_serf__lock(svn_ra_session_t *ra_session,
 
       svn_ra_serf__request_create(handler);
       err = svn_ra_serf__context_run_wait(&lock_ctx->done, session, subpool);
-      if (lock_ctx->error || parser_ctx->error)
-        {
-          svn_error_clear(err);
-          SVN_ERR(lock_ctx->error);
-          SVN_ERR(parser_ctx->error);
-        }
-      if (err)
-        {
-          return svn_error_create(SVN_ERR_RA_DAV_REQUEST_FAILED, err,
-                                  _("Lock request failed"));
-        }
 
-      SVN_ERR(lock_func(lock_baton, lock_ctx->path, TRUE, lock_ctx->lock, NULL,
-                        subpool));
+      if (lock_func)
+        new_err = lock_func(lock_baton, lock_ctx->path, TRUE, lock_ctx->lock,
+                            err, subpool);
+      svn_error_clear(err);
+
+      SVN_ERR(new_err);
     }
 
   return SVN_NO_ERROR;
@@ -652,7 +652,7 @@ struct unlock_context_t {
   svn_boolean_t force;
 };
 
-static apr_status_t
+static svn_error_t *
 set_unlock_headers(serf_bucket_t *headers,
                    void *baton,
                    apr_pool_t *pool)
@@ -666,7 +666,7 @@ set_unlock_headers(serf_bucket_t *headers,
                               SVN_DAV_OPTION_LOCK_BREAK);
     }
 
-  return APR_SUCCESS;
+  return SVN_NO_ERROR;
 }
 
 svn_error_t *
@@ -681,7 +681,7 @@ svn_ra_serf__unlock(svn_ra_session_t *ra_session,
   apr_hash_index_t *hi;
   apr_pool_t *subpool;
 
-  apr_pool_create(&subpool, pool);
+  subpool = svn_pool_create(pool);
 
   for (hi = apr_hash_first(pool, path_tokens); hi; hi = apr_hash_next(hi))
     {
@@ -690,22 +690,24 @@ svn_ra_serf__unlock(svn_ra_session_t *ra_session,
       const char *req_url, *path, *token;
       const void *key;
       void *val;
+      svn_lock_t *existing_lock = NULL;
       struct unlock_context_t unlock_ctx;
+      svn_error_t *lock_err = NULL;
 
       svn_pool_clear(subpool);
 
       ctx = apr_pcalloc(subpool, sizeof(*ctx));
+      ctx->pool = subpool;
 
       apr_hash_this(hi, &key, NULL, &val);
       path = key;
       token = val;
 
-      if (force == TRUE && (!token || token[0] == '\0'))
+      if (force && (!token || token[0] == '\0'))
         {
-          svn_lock_t *lock;
-
-          SVN_ERR(svn_ra_serf__get_lock(ra_session, &lock, path, subpool));
-          token = lock->token;
+          SVN_ERR(svn_ra_serf__get_lock(ra_session, &existing_lock, path,
+                                        subpool));
+          token = existing_lock->token;
           if (!token)
             {
               svn_error_t *err;
@@ -727,10 +729,10 @@ svn_ra_serf__unlock(svn_ra_session_t *ra_session,
         }
 
       unlock_ctx.force = force;
-      unlock_ctx.token = apr_pstrcat(subpool, "<", token, ">", NULL);
+      unlock_ctx.token = apr_pstrcat(subpool, "<", token, ">", (char *)NULL);
 
-      req_url = svn_path_url_add_component(session->repos_url.path, path,
-                                           subpool);
+      req_url = svn_path_url_add_component2(session->session_url.path, path,
+                                            subpool);
 
       handler = apr_pcalloc(subpool, sizeof(*handler));
 
@@ -748,15 +750,28 @@ svn_ra_serf__unlock(svn_ra_session_t *ra_session,
       svn_ra_serf__request_create(handler);
       SVN_ERR(svn_ra_serf__context_run_wait(&ctx->done, session, subpool));
 
-      if (ctx->status != 204)
+      switch (ctx->status)
         {
-           return svn_error_createf(SVN_ERR_RA_DAV_REQUEST_FAILED, NULL,
-                                    _("Unlock request failed: %d %s"),
-                                    ctx->status, ctx->reason);
+          case 204:
+            break; /* OK */
+          case 403:
+            /* Api users expect this specific error code to detect failures */
+            lock_err = svn_error_createf(SVN_ERR_FS_LOCK_OWNER_MISMATCH, NULL,
+                                         _("Unlock request failed: %d %s"),
+                                         ctx->status, ctx->reason);
+            break;
+          default:
+            lock_err = svn_error_createf(SVN_ERR_RA_DAV_REQUEST_FAILED, NULL,
+                                   _("Unlock request failed: %d %s"),
+                                   ctx->status, ctx->reason);
         }
 
       if (lock_func)
-        SVN_ERR(lock_func(lock_baton, path, FALSE, NULL, NULL, subpool));
+        {
+          SVN_ERR(lock_func(lock_baton, path, FALSE, existing_lock,
+                            lock_err, subpool));
+          svn_error_clear(lock_err);
+        }
     }
 
   return SVN_NO_ERROR;
