@@ -2,17 +2,22 @@
  * replay.c :  entry point for replay RA functions for ra_serf
  *
  * ====================================================================
- * Copyright (c) 2006-2007 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and replays, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
@@ -32,7 +37,6 @@
 #include "svn_config.h"
 #include "svn_delta.h"
 #include "svn_base64.h"
-#include "svn_version.h"
 #include "svn_path.h"
 #include "svn_private_config.h"
 
@@ -42,7 +46,7 @@
 /*
  * This enum represents the current state of our XML parsing.
  */
-typedef enum {
+typedef enum replay_state_e {
   NONE = 0,
   REPORT,
   OPEN_DIR,
@@ -71,7 +75,7 @@ typedef svn_error_t *
                  const svn_string_t *value,
                  apr_pool_t *pool);
 
-typedef struct {
+typedef struct prop_info_t {
   apr_pool_t *pool;
 
   change_prop_t change;
@@ -85,8 +89,11 @@ typedef struct {
   replay_info_t *parent;
 } prop_info_t;
 
-typedef struct {
-  apr_pool_t *pool;
+typedef struct replay_context_t {
+  apr_pool_t *src_rev_pool;
+  apr_pool_t *dst_rev_pool;
+  /*file_pool is cleared after completion of each file. */
+  apr_pool_t *file_pool;
 
   /* Are we done fetching this file? */
   svn_boolean_t done;
@@ -109,8 +116,12 @@ typedef struct {
   svn_revnum_t low_water_mark;
   svn_boolean_t send_deltas;
 
-  /* Cached vcc_url */
-  const char *vcc_url;
+  /* Cached report target url */
+  const char *report_target;
+
+  /* Target and revision to fetch revision properties on */
+  const char *revprop_target;
+  svn_revnum_t revprop_rev;
 
   /* Revision properties for this revision. */
   apr_hash_t *revs_props;
@@ -119,6 +130,9 @@ typedef struct {
   /* Keep a reference to the XML parser ctx to report any errors. */
   svn_ra_serf__xml_parser_t *parser_ctx;
 
+  /* The propfind for the revision properties of the current revision */
+  svn_ra_serf__propfind_context_t *prop_ctx;
+
 } replay_context_t;
 
 
@@ -134,9 +148,9 @@ push_state(svn_ra_serf__xml_parser_t *parser,
     {
       replay_info_t *info;
 
-      info = apr_palloc(parser->state->pool, sizeof(*info));
+      info = apr_palloc(replay_ctx->dst_rev_pool, sizeof(*info));
 
-      info->pool = parser->state->pool;
+      info->pool = replay_ctx->dst_rev_pool;
       info->parent = parser->state->private;
       info->baton = NULL;
       info->stream = NULL;
@@ -147,9 +161,9 @@ push_state(svn_ra_serf__xml_parser_t *parser,
     {
       prop_info_t *info;
 
-      info = apr_pcalloc(parser->state->pool, sizeof(*info));
+      info = apr_pcalloc(replay_ctx->dst_rev_pool, sizeof(*info));
 
-      info->pool = parser->state->pool;
+      info->pool = replay_ctx->dst_rev_pool;
       info->parent = parser->state->private;
 
       parser->state->private = info;
@@ -173,17 +187,29 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
       strcmp(name.name, "editor-report") == 0)
     {
       push_state(parser, ctx, REPORT);
-      ctx->props = apr_hash_make(ctx->pool);
 
-      svn_ra_serf__walk_all_props(ctx->revs_props, ctx->vcc_url, ctx->revision, 
-                                  svn_ra_serf__set_bare_props,
-                                  ctx->props, ctx->pool);
+      /* Before we can continue, we need the revision properties. */
+      SVN_ERR_ASSERT(!ctx->prop_ctx
+                     || svn_ra_serf__propfind_is_done(ctx->prop_ctx));
+
+      /* Create a pool for the commit editor. */
+      ctx->dst_rev_pool = svn_pool_create(ctx->src_rev_pool);
+      ctx->file_pool = svn_pool_create(ctx->dst_rev_pool);
+
+      /* ### it would be nice to have a proper scratch_pool.  */
+      SVN_ERR(svn_ra_serf__select_revprops(&ctx->props,
+                                           ctx->revprop_target,
+                                           ctx->revprop_rev,
+                                           ctx->revs_props,
+                                           ctx->dst_rev_pool,
+                                           ctx->dst_rev_pool));
+
       if (ctx->revstart_func)
         {
           SVN_ERR(ctx->revstart_func(ctx->revision, ctx->replay_baton,
                                      &ctx->editor, &ctx->editor_baton,
                                      ctx->props,
-                                     ctx->pool));
+                                     ctx->dst_rev_pool));
         }
     }
   else if (state == REPORT &&
@@ -200,7 +226,7 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
 
       SVN_ERR(ctx->editor->set_target_revision(ctx->editor_baton,
                                                SVN_STR_TO_REV(rev),
-                                               parser->state->pool));
+                                               ctx->dst_rev_pool));
     }
   else if (state == REPORT &&
            strcmp(name.name, "open-root") == 0)
@@ -219,7 +245,8 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
       info = push_state(parser, ctx, OPEN_DIR);
 
       SVN_ERR(ctx->editor->open_root(ctx->editor_baton,
-                                     SVN_STR_TO_REV(rev), parser->state->pool,
+                                     SVN_STR_TO_REV(rev),
+                                     ctx->dst_rev_pool,
                                      &info->baton));
     }
   else if ((state == OPEN_DIR || state == ADD_DIR) &&
@@ -244,7 +271,7 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
       info = push_state(parser, ctx, DELETE_ENTRY);
 
       SVN_ERR(ctx->editor->delete_entry(file_name, SVN_STR_TO_REV(rev),
-                                        info->baton, parser->state->pool));
+                                        info->baton, ctx->dst_rev_pool));
 
       svn_ra_serf__xml_pop_state(parser);
     }
@@ -271,7 +298,7 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
 
       SVN_ERR(ctx->editor->open_directory(dir_name, info->parent->baton,
                                           SVN_STR_TO_REV(rev),
-                                          parser->state->pool, &info->baton));
+                                          ctx->dst_rev_pool, &info->baton));
     }
   else if ((state == OPEN_DIR || state == ADD_DIR) &&
            strcmp(name.name, "add-directory") == 0)
@@ -298,14 +325,14 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
 
       SVN_ERR(ctx->editor->add_directory(dir_name, info->parent->baton,
                                          copyfrom, rev,
-                                         parser->state->pool, &info->baton));
+                                         ctx->dst_rev_pool, &info->baton));
     }
   else if ((state == OPEN_DIR || state == ADD_DIR) &&
            strcmp(name.name, "close-directory") == 0)
     {
       replay_info_t *info = parser->state->private;
 
-      SVN_ERR(ctx->editor->close_directory(info->baton, parser->state->pool));
+      SVN_ERR(ctx->editor->close_directory(info->baton, ctx->dst_rev_pool));
 
       svn_ra_serf__xml_pop_state(parser);
     }
@@ -315,6 +342,7 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
       const char *file_name, *rev;
       replay_info_t *info;
 
+      svn_pool_clear(ctx->file_pool);
       file_name = svn_xml_get_attr_value("name", attrs);
       if (!file_name)
         {
@@ -332,7 +360,7 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
 
       SVN_ERR(ctx->editor->open_file(file_name, info->parent->baton,
                                      SVN_STR_TO_REV(rev),
-                                     parser->state->pool, &info->baton));
+                                     ctx->file_pool, &info->baton));
     }
   else if ((state == OPEN_DIR || state == ADD_DIR) &&
            strcmp(name.name, "add-file") == 0)
@@ -341,6 +369,7 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
       svn_revnum_t rev;
       replay_info_t *info;
 
+      svn_pool_clear(ctx->file_pool);
       file_name = svn_xml_get_attr_value("name", attrs);
       if (!file_name)
         {
@@ -359,7 +388,7 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
 
       SVN_ERR(ctx->editor->add_file(file_name, info->parent->baton,
                                     copyfrom, rev,
-                                    parser->state->pool, &info->baton));
+                                    ctx->file_pool, &info->baton));
     }
   else if ((state == OPEN_FILE || state == ADD_FILE) &&
            strcmp(name.name, "apply-textdelta") == 0)
@@ -379,7 +408,7 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
         }
 
       SVN_ERR(ctx->editor->apply_textdelta(info->baton, checksum,
-                                           info->pool,
+                                           ctx->file_pool,
                                            &textdelta,
                                            &textdelta_baton));
 
@@ -396,7 +425,7 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
       checksum = svn_xml_get_attr_value("checksum", attrs);
 
       SVN_ERR(ctx->editor->close_file(info->baton, checksum,
-                                      parser->state->pool));
+                                      ctx->file_pool));
 
       svn_ra_serf__xml_pop_state(parser);
     }
@@ -418,7 +447,6 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
 
       info = push_state(parser, ctx, CHANGE_PROP);
 
-      info->name = apr_pstrdup(parser->state->pool, prop_name);
 
       if (svn_xml_get_attr_value("del", attrs))
         info->del_prop = TRUE;
@@ -426,9 +454,15 @@ start_replay(svn_ra_serf__xml_parser_t *parser,
         info->del_prop = FALSE;
 
       if (state == OPEN_FILE || state == ADD_FILE)
-        info->change = ctx->editor->change_file_prop;
+        {
+          info->name = apr_pstrdup(ctx->file_pool, prop_name);
+          info->change = ctx->editor->change_file_prop;
+        }
       else
-        info->change = ctx->editor->change_dir_prop;
+        {
+          info->name = apr_pstrdup(ctx->dst_rev_pool, prop_name);
+          info->change = ctx->editor->change_dir_prop;
+        }
 
     }
 
@@ -456,8 +490,9 @@ end_replay(svn_ra_serf__xml_parser_t *parser,
           SVN_ERR(ctx->revfinish_func(ctx->revision, ctx->replay_baton,
                                       ctx->editor, ctx->editor_baton,
                                       ctx->props,
-                                      ctx->pool));
+                                      ctx->dst_rev_pool));
         }
+      svn_pool_destroy(ctx->dst_rev_pool);
     }
   else if (state == OPEN_DIR && strcmp(name.name, "open-directory") == 0)
     {
@@ -494,7 +529,7 @@ end_replay(svn_ra_serf__xml_parser_t *parser,
       prop_info_t *info = parser->state->private;
       const svn_string_t *prop_val;
 
-      if (info->del_prop == TRUE)
+      if (info->del_prop)
         {
           prop_val = NULL;
         }
@@ -505,7 +540,10 @@ end_replay(svn_ra_serf__xml_parser_t *parser,
           tmp_prop.data = info->data;
           tmp_prop.len = info->len;
 
-          prop_val = svn_base64_decode_string(&tmp_prop, parser->state->pool);
+          if (strcmp(name.name, "change-file-prop") == 0)
+            prop_val = svn_base64_decode_string(&tmp_prop, ctx->file_pool);
+          else
+            prop_val = svn_base64_decode_string(&tmp_prop, ctx->dst_rev_pool);
         }
 
       SVN_ERR(info->change(info->parent->baton, info->name, prop_val,
@@ -553,50 +591,40 @@ cdata_replay(svn_ra_serf__xml_parser_t *parser,
   return SVN_NO_ERROR;
 }
 
-static serf_bucket_t *
-create_replay_body(void *baton,
+static svn_error_t *
+create_replay_body(serf_bucket_t **bkt,
+                   void *baton,
                    serf_bucket_alloc_t *alloc,
                    apr_pool_t *pool)
 {
   replay_context_t *ctx = baton;
-  serf_bucket_t *body_bkt, *tmp;
+  serf_bucket_t *body_bkt;
 
   body_bkt = serf_bucket_aggregate_create(alloc);
 
-  tmp = SERF_BUCKET_SIMPLE_STRING_LEN("<S:replay-report xmlns:S=\"",
-                                      sizeof("<S:replay-report xmlns:S=\"")-1,
-                                      alloc);
-  serf_bucket_aggregate_append(body_bkt, tmp);
-
-  tmp = SERF_BUCKET_SIMPLE_STRING_LEN(SVN_XML_NAMESPACE,
-                                      sizeof(SVN_XML_NAMESPACE)-1,
-                                      alloc);
-  serf_bucket_aggregate_append(body_bkt, tmp);
-
-  tmp = SERF_BUCKET_SIMPLE_STRING_LEN("\">",
-                                      sizeof("\">")-1,
-                                      alloc);
-  serf_bucket_aggregate_append(body_bkt, tmp);
+  svn_ra_serf__add_open_tag_buckets(body_bkt, alloc,
+                                    "S:replay-report",
+                                    "xmlns:S", SVN_XML_NAMESPACE,
+                                    NULL);
 
   svn_ra_serf__add_tag_buckets(body_bkt,
-                               "S:revision", apr_ltoa(ctx->pool, ctx->revision),
+                               "S:revision",
+                               apr_ltoa(ctx->src_rev_pool, ctx->revision),
                                alloc);
   svn_ra_serf__add_tag_buckets(body_bkt,
                                "S:low-water-mark",
-                               apr_ltoa(ctx->pool, ctx->low_water_mark),
+                               apr_ltoa(ctx->src_rev_pool, ctx->low_water_mark),
                                alloc);
 
   svn_ra_serf__add_tag_buckets(body_bkt,
                                "S:send-deltas",
-                               apr_ltoa(ctx->pool, ctx->send_deltas),
+                               apr_ltoa(ctx->src_rev_pool, ctx->send_deltas),
                                alloc);
 
-  tmp = SERF_BUCKET_SIMPLE_STRING_LEN("</S:replay-report>",
-                                      sizeof("</S:replay-report>")-1,
-                                      alloc);
-  serf_bucket_aggregate_append(body_bkt, tmp);
+  svn_ra_serf__add_close_tag_buckets(body_bkt, alloc, "S:replay-report");
 
-  return body_bkt;
+  *bkt = body_bkt;
+  return SVN_NO_ERROR;
 }
 
 svn_error_t *
@@ -613,31 +641,29 @@ svn_ra_serf__replay(svn_ra_session_t *ra_session,
   svn_ra_serf__handler_t *handler;
   svn_ra_serf__xml_parser_t *parser_ctx;
   svn_error_t *err;
+  const char *report_target;
   /* We're not really interested in the status code here in replay, but
      the XML parsing code will abort on error if it doesn't have a place
      to store the response status code. */
   int status_code;
-  const char *vcc_url;
 
-  SVN_ERR(svn_ra_serf__discover_root(&vcc_url, NULL,
-                                     session, session->conns[0],
-                                     session->repos_url.path, pool));
+  SVN_ERR(svn_ra_serf__report_resource(&report_target, session, NULL, pool));
 
   replay_ctx = apr_pcalloc(pool, sizeof(*replay_ctx));
-  replay_ctx->pool = pool;
+  replay_ctx->src_rev_pool = pool;
   replay_ctx->editor = editor;
   replay_ctx->editor_baton = edit_baton;
   replay_ctx->done = FALSE;
   replay_ctx->revision = revision;
   replay_ctx->low_water_mark = low_water_mark;
   replay_ctx->send_deltas = send_deltas;
-  replay_ctx->vcc_url = vcc_url;
-  replay_ctx->revs_props = apr_hash_make(replay_ctx->pool);
+  replay_ctx->report_target = report_target;
+  replay_ctx->revs_props = apr_hash_make(replay_ctx->src_rev_pool);
 
   handler = apr_pcalloc(pool, sizeof(*handler));
 
   handler->method = "REPORT";
-  handler->path = session->repos_url_str;
+  handler->path = session->session_url_str;
   handler->body_delegate = create_replay_body;
   handler->body_delegate_baton = replay_ctx;
   handler->body_type = "text/xml";
@@ -664,40 +690,39 @@ svn_ra_serf__replay(svn_ra_session_t *ra_session,
 
   err = svn_ra_serf__context_run_wait(&replay_ctx->done, session, pool);
 
-  if (parser_ctx->error) {
-    svn_error_clear(err);
-    return parser_ctx->error;
-  }
   SVN_ERR(err);
 
   return SVN_NO_ERROR;
 }
 
-/* The maximum number of outstanding requests at any time. When this number is
- * reached, ra_serf will stop sending requests until responses on the previous
- * requests are received and handled.
+/* The maximum number of outstanding requests at any time. When this
+ * number is reached, ra_serf will stop sending requests until
+ * responses on the previous requests are received and handled.
  *
  * Some observations about serf which lead us to the current value.
  * ----------------------------------------------------------------
- * We aim to keep serf's outgoing queue filled with enough requests so the 
- * network bandwidth and server capacity is used optimally. Originally we used
- * 5 as the max. number of outstanding requests, but this turned out to be too
- * low. 
- * Serf doesn't exit out of the serf_context_run loop as long as it has 
- * data to send or receive. With small responses (revs of a few kB), serf 
- * doesn't come out of this loop at all. So with MAX_OUTSTANDING_REQUESTS set
- * to a low number, there's a big chance that serf handles those requests
- * completely in its internal loop, and only then gives us a chance to create
- * new requests. This results in hiccups, slowing down the whole process.
  *
- * With a larger MAX_OUTSTANDING_REQUESTS, like 100 or more, there's more chance
- * that serf can come out of its internal loop so we can replenish the outgoing
- * request queue.
- * There's no real disadvantage of using a large number here, besides the memory
- * used to store the message, parser and handler objects (approx. 250 bytes). 
- * 
- * In my test setup peak performance was reached at max. 30-35 requests. So I
- * added a small margin and choose 50.
+ * We aim to keep serf's outgoing queue filled with enough requests so
+ * the network bandwidth and server capacity is used
+ * optimally. Originally we used 5 as the max. number of outstanding
+ * requests, but this turned out to be too low.
+ *
+ * Serf doesn't exit out of the serf_context_run loop as long as it
+ * has data to send or receive. With small responses (revs of a few
+ * kB), serf doesn't come out of this loop at all. So with
+ * MAX_OUTSTANDING_REQUESTS set to a low number, there's a big chance
+ * that serf handles those requests completely in its internal loop,
+ * and only then gives us a chance to create new requests. This
+ * results in hiccups, slowing down the whole process.
+ *
+ * With a larger MAX_OUTSTANDING_REQUESTS, like 100 or more, there's
+ * more chance that serf can come out of its internal loop so we can
+ * replenish the outgoing request queue.  There's no real disadvantage
+ * of using a large number here, besides the memory used to store the
+ * message, parser and handler objects (approx. 250 bytes).
+ *
+ * In my test setup peak performance was reached at max. 30-35
+ * requests. So I added a small margin and chose 50.
  */
 #define MAX_OUTSTANDING_REQUESTS 50
 
@@ -714,32 +739,36 @@ svn_ra_serf__replay_range(svn_ra_session_t *ra_session,
 {
   svn_ra_serf__session_t *session = ra_session->priv;
   svn_revnum_t rev = start_revision;
-  const char *vcc_url;
+  const char *report_target;
   int active_reports = 0;
 
-  SVN_ERR(svn_ra_serf__discover_root(&vcc_url, NULL,
-                                     session, session->conns[0],
-                                     session->repos_url.path, pool));
+  SVN_ERR(svn_ra_serf__report_resource(&report_target, session, NULL, pool));
 
   while (active_reports || rev <= end_revision)
     {
       apr_status_t status;
+      svn_error_t *err;
       svn_ra_serf__list_t *done_list;
       svn_ra_serf__list_t *done_reports = NULL;
       replay_context_t *replay_ctx;
+      /* We're not really interested in the status code here in replay, but
+         the XML parsing code will abort on error if it doesn't have a place
+         to store the response status code. */
       int status_code;
 
-      /* Send pending requests, if any. Limit the number of outstanding 
+      if (session->cancel_func)
+        SVN_ERR(session->cancel_func(session->cancel_baton));
+
+      /* Send pending requests, if any. Limit the number of outstanding
          requests to MAX_OUTSTANDING_REQUESTS. */
       if (rev <= end_revision  && active_reports < MAX_OUTSTANDING_REQUESTS)
         {
-          svn_ra_serf__propfind_context_t *prop_ctx = NULL;
           svn_ra_serf__handler_t *handler;
           svn_ra_serf__xml_parser_t *parser_ctx;
           apr_pool_t *ctx_pool = svn_pool_create(pool);
 
           replay_ctx = apr_pcalloc(ctx_pool, sizeof(*replay_ctx));
-          replay_ctx->pool = ctx_pool;
+          replay_ctx->src_rev_pool = ctx_pool;
           replay_ctx->revstart_func = revstart_func;
           replay_ctx->revfinish_func = revfinish_func;
           replay_ctx->replay_baton = replay_baton;
@@ -749,34 +778,51 @@ svn_ra_serf__replay_range(svn_ra_session_t *ra_session,
           replay_ctx->send_deltas = send_deltas;
           replay_ctx->done_item.data = replay_ctx;
           /* Request all properties of a certain revision. */
-          replay_ctx->vcc_url = vcc_url;
-          replay_ctx->revs_props = apr_hash_make(replay_ctx->pool);
-          SVN_ERR(svn_ra_serf__deliver_props(&prop_ctx, 
+          replay_ctx->report_target = report_target;
+          replay_ctx->revs_props = apr_hash_make(replay_ctx->src_rev_pool);
+
+          if (SVN_RA_SERF__HAVE_HTTPV2_SUPPORT(session))
+           {
+             replay_ctx->revprop_target = apr_psprintf(pool, "%s/%ld",
+                                                       session->rev_stub, rev);
+             replay_ctx->revprop_rev = SVN_INVALID_REVNUM;
+            }
+          else
+            {
+              replay_ctx->revprop_target = report_target;
+              replay_ctx->revprop_rev = rev;
+            }
+
+          SVN_ERR(svn_ra_serf__deliver_props(&replay_ctx->prop_ctx,
                                              replay_ctx->revs_props, session,
-                                             session->conns[0], vcc_url,
-                                             rev,  "0", all_props,
-                                             TRUE, NULL, replay_ctx->pool));
+                                             session->conns[0],
+                                             replay_ctx->revprop_target,
+                                             replay_ctx->revprop_rev,
+                                             "0", all_props,
+                                             NULL,
+                                             replay_ctx->src_rev_pool));
 
           /* Send the replay report request. */
-          handler = apr_pcalloc(replay_ctx->pool, sizeof(*handler));
+          handler = apr_pcalloc(replay_ctx->src_rev_pool, sizeof(*handler));
 
           handler->method = "REPORT";
-          handler->path = session->repos_url_str;
+          handler->path = session->session_url_str;
           handler->body_delegate = create_replay_body;
           handler->body_delegate_baton = replay_ctx;
           handler->conn = session->conns[0];
           handler->session = session;
 
-          parser_ctx = apr_pcalloc(replay_ctx->pool, sizeof(*parser_ctx));
+          parser_ctx = apr_pcalloc(replay_ctx->src_rev_pool,
+                                   sizeof(*parser_ctx));
 
           /* Setup the XML parser context.
              Because we have not one but a list of requests, the 'done' property
-             on the replay_ctx is not of much use. Instead, use 'done_list'. 
+             on the replay_ctx is not of much use. Instead, use 'done_list'.
              On each handled response (succesfully or not), the parser will add
-             done_item to done_list, so by keeping track of the state of 
-             done_list we know how many requests have been handled completely. 
+             done_item to done_list, so by keeping track of the state of
+             done_list we know how many requests have been handled completely.
           */
-          parser_ctx->pool = replay_ctx->pool;
+          parser_ctx->pool = replay_ctx->src_rev_pool;
           parser_ctx->user_data = replay_ctx;
           parser_ctx->start = start_replay;
           parser_ctx->end = end_replay;
@@ -797,37 +843,40 @@ svn_ra_serf__replay_range(svn_ra_session_t *ra_session,
           active_reports++;
         }
 
-      /* Run the serf loop, send outgoing and process incoming requests. 
-         This request will block when there are no more requests to send or 
+      /* Run the serf loop, send outgoing and process incoming requests.
+         This request will block when there are no more requests to send or
          responses to receive, so we have to be careful on our bookkeeping. */
-      status = serf_context_run(session->context, SERF_DURATION_FOREVER, 
+      status = serf_context_run(session->context, session->timeout,
                                 pool);
 
-      /* Substract the number of completely handled responses from our 
+      err = session->pending_error;
+      session->pending_error = NULL;
+
+      if (APR_STATUS_IS_TIMEUP(status))
+        {
+          svn_error_clear(err);
+          return svn_error_create(SVN_ERR_RA_DAV_CONN_TIMEOUT,
+                                  NULL,
+                                  _("Connection timed out"));
+        }
+
+      /* Substract the number of completely handled responses from our
          total nr. of open requests', so we'll know when to stop this loop.
          Since the message is completely handled, we can destroy its pool. */
       done_list = done_reports;
       while (done_list)
         {
           replay_context_t *ctx = (replay_context_t *)done_list->data;
-          svn_ra_serf__xml_parser_t *parser_ctx = ctx->parser_ctx;
-          if (parser_ctx->error)
-            {
-              svn_error_clear(session->pending_error);
-              session->pending_error = SVN_NO_ERROR;
-              SVN_ERR(parser_ctx->error);
-            }
 
           done_list = done_list->next;
-          svn_pool_destroy(ctx->pool);
+          svn_pool_destroy(ctx->src_rev_pool);
           active_reports--;
         }
 
+      SVN_ERR(err);
       if (status)
         {
-          SVN_ERR(session->pending_error);
-
-          return svn_error_wrap_apr(status, 
+          return svn_error_wrap_apr(status,
                                     _("Error retrieving replay REPORT (%d)"),
                                     status);
         }
