@@ -1,34 +1,39 @@
 /* authz.c : path-based access control
  *
  * ====================================================================
- * Copyright (c) 2000-2006 CollabNet.  All rights reserved.
+ *    Licensed to the Apache Software Foundation (ASF) under one
+ *    or more contributor license agreements.  See the NOTICE file
+ *    distributed with this work for additional information
+ *    regarding copyright ownership.  The ASF licenses this file
+ *    to you under the Apache License, Version 2.0 (the
+ *    "License"); you may not use this file except in compliance
+ *    with the License.  You may obtain a copy of the License at
  *
- * This software is licensed as described in the file COPYING, which
- * you should have received as part of this distribution.  The terms
- * are also available at http://subversion.tigris.org/license-1.html.
- * If newer versions of this license are posted there, you may use a
- * newer version instead, at your option.
+ *      http://www.apache.org/licenses/LICENSE-2.0
  *
- * This software consists of voluntary contributions made by many
- * individuals.  For exact contribution history, see the revision
- * history and logs, available at http://subversion.tigris.org/.
+ *    Unless required by applicable law or agreed to in writing,
+ *    software distributed under the License is distributed on an
+ *    "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+ *    KIND, either express or implied.  See the License for the
+ *    specific language governing permissions and limitations
+ *    under the License.
  * ====================================================================
  */
 
 
 /*** Includes. ***/
 
-#include <assert.h>
-
 #include <apr_pools.h>
 #include <apr_file_io.h>
 
 #include "svn_pools.h"
 #include "svn_error.h"
+#include "svn_dirent_uri.h"
 #include "svn_path.h"
 #include "svn_repos.h"
 #include "svn_config.h"
 #include "svn_ctype.h"
+#include "private/svn_fspath.h"
 
 
 /*** Structures. ***/
@@ -52,9 +57,9 @@ struct authz_lookup_baton {
 
   /* The following are used exclusively in recursive lookups. */
 
-  /* The path in the repository to authorize. */
+  /* The path in the repository (an fspath) to authorize. */
   const char *repos_path;
-  /* repos_path prefixed by the repository name. */
+  /* repos_path prefixed by the repository name and a colon. */
   const char *qualified_repos_path;
 
   /* Whether, at the end of a recursive lookup, access is granted. */
@@ -123,6 +128,28 @@ authz_access_is_determined(svn_repos_authz_access_t allow,
     return FALSE;
 }
 
+/* Return TRUE is USER equals ALIAS. The alias definitions are in the
+   "aliases" sections of CFG. Use POOL for temporary allocations during
+   the lookup. */
+static svn_boolean_t
+authz_alias_is_user(svn_config_t *cfg,
+                    const char *alias,
+                    const char *user,
+                    apr_pool_t *pool)
+{
+  const char *value;
+
+  svn_config_get(cfg, &value, "aliases", alias, NULL);
+  if (!value)
+    return FALSE;
+
+  if (strcmp(value, user) == 0)
+    return TRUE;
+
+  return FALSE;
+}
+
+
 /* Return TRUE if USER is in GROUP.  The group definitions are in the
    "groups" section of CFG.  Use POOL for temporary allocations during
    the lookup. */
@@ -152,6 +179,14 @@ authz_group_contains_user(svn_config_t *cfg,
             return TRUE;
         }
 
+      /* If the 'user' is an alias, verify it. */
+      else if (*group_user == '&')
+        {
+          if (authz_alias_is_user(cfg, &group_user[1],
+                                  user, pool))
+            return TRUE;
+        }
+
       /* If the user matches, stop. */
       else if (strcmp(user, group_user) == 0)
         return TRUE;
@@ -161,34 +196,70 @@ authz_group_contains_user(svn_config_t *cfg,
 }
 
 
+/* Determines whether an authz rule applies to the current
+ * user, given the name part of the rule's name-value pair
+ * in RULE_MATCH_STRING and the authz_lookup_baton object
+ * B with the username in question.
+ */
+static svn_boolean_t
+authz_line_applies_to_user(const char *rule_match_string,
+                           struct authz_lookup_baton *b,
+                           apr_pool_t *pool)
+{
+  /* If the rule has an inversion, recurse and invert the result. */
+  if (rule_match_string[0] == '~')
+    return !authz_line_applies_to_user(&rule_match_string[1], b, pool);
+
+  /* Check for special tokens. */
+  if (strcmp(rule_match_string, "$anonymous") == 0)
+    return (b->user == NULL);
+  if (strcmp(rule_match_string, "$authenticated") == 0)
+    return (b->user != NULL);
+
+  /* Check for a wildcard rule. */
+  if (strcmp(rule_match_string, "*") == 0)
+    return TRUE;
+
+  /* If we get here, then the rule is:
+   *  - Not an inversion rule.
+   *  - Not an authz token rule.
+   *  - Not a wildcard rule.
+   *
+   * All that's left over is regular user or group specifications.
+   */
+
+  /* If the session is anonymous, then a user/group
+   * rule definitely won't match.
+   */
+  if (b->user == NULL)
+    return FALSE;
+
+  /* Process the rule depending on whether it is
+   * a user, alias or group rule.
+   */
+  if (rule_match_string[0] == '@')
+    return authz_group_contains_user(
+      b->config, &rule_match_string[1], b->user, pool);
+  else if (rule_match_string[0] == '&')
+    return authz_alias_is_user(
+      b->config, &rule_match_string[1], b->user, pool);
+  else
+    return (strcmp(b->user, rule_match_string) == 0);
+}
+
+
 /* Callback to parse one line of an authz file and update the
  * authz_baton accordingly.
  */
 static svn_boolean_t
-authz_parse_line(const char *name, const char *value, 
+authz_parse_line(const char *name, const char *value,
                  void *baton, apr_pool_t *pool)
 {
   struct authz_lookup_baton *b = baton;
 
-  /* Work out whether this ACL line applies to the user. */
-  if (strcmp(name, "*") != 0)
-    {
-      /* Non-anon rule, anon user.  Stop. */
-      if (!b->user)
-        return TRUE;
-
-      /* Group rule and user not in group.  Stop. */
-      if (*name == '@')
-        {
-          if (!authz_group_contains_user(b->config, &name[1],
-                                         b->user, pool))
-            return TRUE;
-        }
-
-      /* User rule for wrong user.  Stop. */
-      else if (strcmp(name, b->user) != 0)
-        return TRUE;
-    }
+  /* Stop if the rule doesn't apply to this user. */
+  if (!authz_line_applies_to_user(name, b, pool))
+    return TRUE;
 
   /* Set the access grants for the rule. */
   if (strchr(value, 'r'))
@@ -205,6 +276,23 @@ authz_parse_line(const char *name, const char *value,
 }
 
 
+/* Return TRUE iff the access rules in SECTION_NAME apply to PATH_SPEC
+ * (which is a repository name, colon, and repository fspath, such as
+ * "myrepos:/trunk/foo").
+ */
+static svn_boolean_t
+is_applicable_section(const char *path_spec,
+                      const char *section_name)
+{
+  apr_size_t path_spec_len = strlen(path_spec);
+
+  return ((strncmp(path_spec, section_name, path_spec_len) == 0)
+          && (path_spec[path_spec_len - 1] == '/'
+              || section_name[path_spec_len] == '/'
+              || section_name[path_spec_len] == '\0'));
+}
+
+
 /* Callback to parse a section and update the authz_baton if the
  * section denies access to the subtree the baton describes.
  */
@@ -215,10 +303,10 @@ authz_parse_section(const char *section_name, void *baton, apr_pool_t *pool)
   svn_boolean_t conclusive;
 
   /* Does the section apply to us? */
-  if (svn_path_is_ancestor(b->qualified_repos_path,
-                           section_name) == FALSE
-      && svn_path_is_ancestor(b->repos_path,
-                              section_name) == FALSE)
+  if (is_applicable_section(b->qualified_repos_path,
+                            section_name) == FALSE
+      && is_applicable_section(b->repos_path,
+                               section_name) == FALSE)
     return TRUE;
 
   /* Work out what this section grants. */
@@ -263,7 +351,7 @@ authz_get_path_access(svn_config_t *cfg, const char *repos_name,
   baton.user = user;
 
   /* Try to locate a repository-specific block first. */
-  qualified_path = apr_pstrcat(pool, repos_name, ":", path, NULL);
+  qualified_path = apr_pstrcat(pool, repos_name, ":", path, (char *)NULL);
   svn_config_enumerate2(cfg, qualified_path,
                         authz_parse_line, &baton, pool);
 
@@ -306,7 +394,7 @@ authz_get_tree_access(svn_config_t *cfg, const char *repos_name,
   baton.required_access = required_access;
   baton.repos_path = path;
   baton.qualified_repos_path = apr_pstrcat(pool, repos_name,
-                                           ":", path, NULL);
+                                           ":", path, (char *)NULL);
   /* Default to access granted if no rules say otherwise. */
   baton.access = TRUE;
 
@@ -321,15 +409,15 @@ authz_get_tree_access(svn_config_t *cfg, const char *repos_name,
    any kind of granted access.  Implements the
    svn_config_section_enumerator2_t interface. */
 static svn_boolean_t
-authz_global_parse_section(const char *section_name, void *baton,
-                           apr_pool_t *pool)
+authz_get_any_access_parser_cb(const char *section_name, void *baton,
+                               apr_pool_t *pool)
 {
   struct authz_lookup_baton *b = baton;
 
   /* Does the section apply to the query? */
   if (section_name[0] == '/'
-      || strncmp(section_name, b->repos_path,
-                 strlen(b->repos_path)) == 0)
+      || strncmp(section_name, b->qualified_repos_path,
+                 strlen(b->qualified_repos_path)) == 0)
     {
       b->allow = b->deny = svn_authz_none;
 
@@ -352,10 +440,10 @@ authz_global_parse_section(const char *section_name, void *baton,
  * to any path within the REPOSITORY.  Return TRUE if so.  Use POOL
  * for temporary allocations. */
 static svn_boolean_t
-authz_get_global_access(svn_config_t *cfg, const char *repos_name,
-                        const char *user,
-                        svn_repos_authz_access_t required_access,
-                        apr_pool_t *pool)
+authz_get_any_access(svn_config_t *cfg, const char *repos_name,
+                     const char *user,
+                     svn_repos_authz_access_t required_access,
+                     apr_pool_t *pool)
 {
   struct authz_lookup_baton baton = { 0 };
 
@@ -363,9 +451,16 @@ authz_get_global_access(svn_config_t *cfg, const char *repos_name,
   baton.user = user;
   baton.required_access = required_access;
   baton.access = FALSE; /* Deny access by default. */
-  baton.repos_path = apr_pstrcat(pool, repos_name, ":/", NULL);
+  baton.repos_path = "/";
+  baton.qualified_repos_path = apr_pstrcat(pool, repos_name,
+                                           ":/", (char *)NULL);
 
-  svn_config_enumerate_sections2(cfg, authz_global_parse_section,
+  /* We could have used svn_config_enumerate2 for "repos_name:/".
+   * However, this requires access for root explicitly (which the user
+   * may not always have). So we end up enumerating the sections in
+   * the authz CFG and stop on the first match with some access for
+   * this user. */
+  svn_config_enumerate_sections2(cfg, authz_get_any_access_parser_cb,
                                  &baton, pool);
 
   /* If walking the configuration was inconclusive, deny access. */
@@ -440,27 +535,84 @@ authz_group_walk(svn_config_t *cfg,
           apr_hash_set(checked_groups, &group_user[1],
                        APR_HASH_KEY_STRING, NULL);
         }
+      else if (*group_user == '&')
+        {
+          const char *alias;
+
+          svn_config_get(cfg, &alias, "aliases", &group_user[1], NULL);
+          /* Having a non-existent alias in the ACL configuration might be the
+             sign of a typo.  Refuse to perform authz on uncertain rules. */
+          if (!alias)
+            return svn_error_createf(SVN_ERR_AUTHZ_INVALID_CONFIG, NULL,
+                                     "An authz rule refers to alias '%s', "
+                                     "which is undefined",
+                                     &group_user[1]);
+        }
     }
 
   return SVN_NO_ERROR;
 }
 
 
-/* Callback to check whether GROUP is a group name, and if so, whether
-   the group definition exists.  Return TRUE if the rule has no
-   errors.  Use BATON for context and error reporting. */
-static svn_boolean_t authz_validate_rule(const char *group,
+/* Callback to perform some simple sanity checks on an authz rule.
+ *
+ * - If RULE_MATCH_STRING references a group or an alias, verify that
+ *   the group or alias definition exists.
+ * - If RULE_MATCH_STRING specifies a token (starts with $), verify
+ *   that the token name is valid.
+ * - If RULE_MATCH_STRING is using inversion, verify that it isn't
+ *   doing it more than once within the one rule, and that it isn't
+ *   "~*", as that would never match.
+ * - Check that VALUE part of the rule specifies only allowed rule
+ *   flag characters ('r' and 'w').
+ *
+ * Return TRUE if the rule has no errors. Use BATON for context and
+ * error reporting.
+ */
+static svn_boolean_t authz_validate_rule(const char *rule_match_string,
                                          const char *value,
                                          void *baton,
                                          apr_pool_t *pool)
 {
   const char *val;
+  const char *match = rule_match_string;
   struct authz_validate_baton *b = baton;
 
-  /* If the rule applies to a group, check its existence. */
-  if (*group == '@')
+  /* Make sure the user isn't using double-negatives. */
+  if (match[0] == '~')
     {
-      svn_config_get(b->config, &val, "groups", &group[1], NULL);
+      /* Bump the pointer past the inversion for the other checks. */
+      match++;
+
+      /* Another inversion is a double negative; we can't not stop. */
+      if (match[0] == '~')
+        {
+          b->err = svn_error_createf(SVN_ERR_AUTHZ_INVALID_CONFIG, NULL,
+                                     "Rule '%s' has more than one "
+                                     "inversion; double negatives are "
+                                     "not permitted.",
+                                     rule_match_string);
+          return FALSE;
+        }
+
+      /* Make sure that the rule isn't "~*", which won't ever match. */
+      if (strcmp(match, "*") == 0)
+        {
+          b->err = svn_error_create(SVN_ERR_AUTHZ_INVALID_CONFIG, NULL,
+                                    "Authz rules with match string '~*' "
+                                    "are not allowed, because they never "
+                                    "match anyone.");
+          return FALSE;
+        }
+    }
+
+  /* If the rule applies to a group, check its existence. */
+  if (match[0] == '@')
+    {
+      const char *group = &match[1];
+
+      svn_config_get(b->config, &val, "groups", group, NULL);
+
       /* Having a non-existent group in the ACL configuration might be
          the sign of a typo.  Refuse to perform authz on uncertain
          rules. */
@@ -469,7 +621,39 @@ static svn_boolean_t authz_validate_rule(const char *group,
           b->err = svn_error_createf(SVN_ERR_AUTHZ_INVALID_CONFIG, NULL,
                                      "An authz rule refers to group "
                                      "'%s', which is undefined",
-                                     group);
+                                     rule_match_string);
+          return FALSE;
+        }
+    }
+
+  /* If the rule applies to an alias, check its existence. */
+  if (match[0] == '&')
+    {
+      const char *alias = &match[1];
+
+      svn_config_get(b->config, &val, "aliases", alias, NULL);
+
+      if (!val)
+        {
+          b->err = svn_error_createf(SVN_ERR_AUTHZ_INVALID_CONFIG, NULL,
+                                     "An authz rule refers to alias "
+                                     "'%s', which is undefined",
+                                     rule_match_string);
+          return FALSE;
+        }
+     }
+
+  /* If the rule specifies a token, check its validity. */
+  if (match[0] == '$')
+    {
+      const char *token_name = &match[1];
+
+      if ((strcmp(token_name, "anonymous") != 0)
+       && (strcmp(token_name, "authenticated") != 0))
+        {
+          b->err = svn_error_createf(SVN_ERR_AUTHZ_INVALID_CONFIG, NULL,
+                                     "Unrecognized authz token '%s'.",
+                                     rule_match_string);
           return FALSE;
         }
     }
@@ -482,7 +666,8 @@ static svn_boolean_t authz_validate_rule(const char *group,
         {
           b->err = svn_error_createf(SVN_ERR_AUTHZ_INVALID_CONFIG, NULL,
                                      "The character '%c' in rule '%s' is not "
-                                     "allowed in authz rules", *val, group);
+                                     "allowed in authz rules", *val,
+                                     rule_match_string);
           return FALSE;
         }
 
@@ -492,6 +677,17 @@ static svn_boolean_t authz_validate_rule(const char *group,
   return TRUE;
 }
 
+/* Callback to check ALIAS's definition for validity.  Use
+   BATON for context and error reporting. */
+static svn_boolean_t authz_validate_alias(const char *alias,
+                                          const char *value,
+                                          void *baton,
+                                          apr_pool_t *pool)
+{
+  /* No checking at the moment, every alias is valid */
+  return TRUE;
+}
+
 
 /* Callback to check GROUP's definition for cyclic dependancies.  Use
    BATON for context and error reporting. */
@@ -523,6 +719,9 @@ static svn_boolean_t authz_validate_section(const char *name,
   if (strncmp(name, "groups", 6) == 0)
     svn_config_enumerate2(b->config, name, authz_validate_group,
                           baton, pool);
+  else if (strncmp(name, "aliases", 7) == 0)
+    svn_config_enumerate2(b->config, name, authz_validate_alias,
+                          baton, pool);
   else
     svn_config_enumerate2(b->config, name, authz_validate_rule,
                           baton, pool);
@@ -547,7 +746,7 @@ svn_repos_authz_read(svn_authz_t **authz_p, const char *file,
   baton.err = SVN_NO_ERROR;
 
   /* Load the rule file. */
-  SVN_ERR(svn_config_read(&authz->cfg, file, must_exist, pool));
+  SVN_ERR(svn_config_read2(&authz->cfg, file, must_exist, TRUE, pool));
   baton.config = authz->cfg;
 
   /* Step through the entire rule file, stopping on error. */
@@ -567,18 +766,23 @@ svn_repos_authz_check_access(svn_authz_t *authz, const char *repos_name,
                              svn_boolean_t *access_granted,
                              apr_pool_t *pool)
 {
-  const char *current_path = path;
+  const char *current_path;
 
-  /* If PATH is NULL, do a global access lookup. */
+  /* If PATH is NULL, check if the user has *any* access. */
   if (!path)
     {
-      *access_granted = authz_get_global_access(authz->cfg, repos_name,
-                                                user, required_access,
-                                                pool);
+      *access_granted = authz_get_any_access(authz->cfg, repos_name,
+                                             user, required_access, pool);
       return SVN_NO_ERROR;
     }
 
+  /* Sanity check. */
+  SVN_ERR_ASSERT(path[0] == '/');
+
   /* Determine the granted access for the requested path. */
+  path = svn_fspath__canonicalize(path, pool);
+  current_path = path;
+
   while (!authz_get_path_access(authz->cfg, repos_name,
                                 current_path, user,
                                 required_access,
@@ -595,7 +799,7 @@ svn_repos_authz_check_access(svn_authz_t *authz, const char *repos_name,
         }
 
       /* Work back to the parent path. */
-      svn_path_split(current_path, &current_path, NULL, pool);
+      current_path = svn_fspath__dirname(current_path, pool);
     }
 
   /* If the caller requested recursive access, we need to walk through
